#include <Servo.h>
#include <IRremote.h>	// See https://decibel.ni.com/content/groups/labview-interface-for-arduino/blog/2014/05/07/irremotetoolscpp5-error-tkd2-was-not-declared-in-this-scope
#include <NewPing.h>	// From https://code.google.com/p/arduino-new-ping/downloads/detail?name=NewPing_v1.5.zip&can=2&q=

/*
	6 legged walking robot code
	Leg Servos connected to pins 2..7
	Echo sensor on servo connected to pin 8
	Ping trigger pin 13
	Ping return pin 12
	IR connected to pin 11
 */

// Left center leg positions
byte const lc_m=100;
byte const lc_l=lc_m+10;
byte const lc_u=lc_m-20;

// Right center leg positions
byte const rc_m=105;
byte const rc_l=rc_m-10;
byte const rc_u=rc_m+20;

// Left front/rear servo mid / front / rear positions
byte const l_m=100;
byte const l_f=l_m-15;
byte const l_r=l_m+15;

// Right front/rear servo mid / front / rear positions
byte const r_m=90;
byte const r_f=r_m+15;
byte const r_r=r_m-15;

Servo slr,slc,slf;	// Left rear,centre,front
Servo srr,src,srf;	// Right rear,centre,front
Servo se;		// Echo sensor left/center/right

void servo_home(void);
decode_results results;
IRrecv irrecv(11);
NewPing sonar(13, 12);

// the setup routine runs once when you press reset:
void setup()
{	// initialize the servo channels
	srr.attach(2);	// Right Rear
	src.attach(3);	// Right Centre
	srf.attach(4);	// Right Front
	slr.attach(5);	// Left Rear
	slc.attach(6);	// Left Centre
	slf.attach(7);	// Left Front
	se.attach(8);	// Echo position
	// Set to midpoint
	se.write(90);
	servo_home();

	// Start the IR receiver
	irrecv.enableIRIn();
	Serial.begin(9600);
}

void servo_pos(byte a,byte b,byte c,byte d,byte e,byte f)
{	slf.write(a);
	slc.write(b);
	slr.write(c);
	srf.write(d);
	src.write(e);
	srr.write(f);
}
	
void servo_home()
{	servo_pos(l_m,lc_m,l_m,r_m,rc_m,r_m);
	se.write(90);
}


// Remember side state
char const s_l= -1;
char const s_m= 0;
char const s_r= 1;
char side=s_m;

void no_lift()
{	side=s_m;
	slc.write(lc_m);
	src.write(rc_m);
	delay(100);
}

// Lift the left side of the robot
void left_lift()
{	if (side!=s_l)
	{	side=s_l;
		delay(200);
		slc.write(lc_l);
		src.write(rc_u);
		delay(200);
	}
}

// Lift the right side of the robot
void right_lift()
{	if (side!=s_r)
	{	side=s_r;
		delay(200);
		slc.write(lc_u);
		src.write(rc_l);
		delay(200);
	}
}


// Move the left legs to "pos"
void left_side(byte posf,byte posr)
{	if (slf.read()!=posf || slr.read()!=posr)
	{	slf.write(posf);
		slr.write(posr);
		delay(20);
	}
}

// Move the left legs from "from" to "to"
void left_slow(byte from,byte to)
{	byte pos;
	if (from<to)
	{	for (pos=from ; pos<=to ; pos+=2)
		{	left_side(pos,pos);
		}
	} else
	{	for (pos=from ; pos>=to ; pos-=2)
		{	left_side(pos,pos);
		}
	}
}

// Move the right legs to "pos"
void right_side(byte posf,byte posr)
{	if (srf.read()!=posf || srr.read()!=posr)
	{	srf.write(posf);
		srr.write(posr);
		delay(20);
	}
}

// Move the right legs from "from" to "to"
void right_slow(byte from,byte to)
{	byte pos;
	if (from<to)
	{	for (pos=from ; pos<=to ; pos+=2)
		{	right_side(pos,pos);
		}
	} else
	{	for (pos=from ; pos>=to ; pos-=2)
		{	right_side(pos,pos);
		}
	}
}

// Do one step forwards
void forward()
{	if (side==s_m)
	{	left_lift();
	}
	if (side==s_l)
	{	left_side(l_f,l_f);
	}
	if (side!=s_r)
	{	right_lift();
	}
	right_side(r_f,r_r);
	left_slow(l_f,l_r);
	right_side(r_f,r_f);
	left_lift();
	left_side(l_f,l_r);
	right_slow(r_f,r_r);
}

// Do one step backwards
void reverse()
{	if (side==s_m)
	{	left_lift();
	}
	if (side==s_l)
	{	left_side(l_r,l_r);
	}
	if (side!=s_r)
	{	right_lift();
	}
	right_side(r_f,r_r);
	left_slow(l_r,l_f);
	right_side(r_r,r_r);
	left_lift();
	left_side(l_f,l_r);
	right_slow(r_r,r_f);
}

// Do one turn left
void left()
{	if (side!=s_r)
	{	right_lift();
	}
	right_side(r_f,r_f);
	left_lift();
	left_side(l_f,l_r);
	right_slow(r_f,r_r);
}

// Do one turn right
void right()
{	if (side!=s_l)
	{	left_lift();
	}
	left_side(l_f,l_f);
	right_lift();
	right_side(r_f,r_r);
	left_slow(l_f,l_r);
}

char key_was_real=0;

int decode_remote()
{	int key=0;
	if (irrecv.decode(&results))
	{	if ((results.value==0xFFFFFFFF && key_was_real==0) ||
			(((results.value>>24)&0xFF)==(((results.value>>16)&0xFF)^0xFF) &&
			((results.value>>8)&0xFF)==((results.value&0xFF)^0xFF)))
		{	key_was_real=1;
			key=((results.value>>16)&0xFF00)|((results.value>>8)&0xFF);
		} else
		{	key_was_real=0;
		}
		irrecv.resume(); // Receive the next value
	}
	return key;
}

int distance[3]= { 999,999,999 };
byte echo_pos=1;

// Scan next position in sequence
void echo_scan()
{	distance[(echo_pos==3) ? 1 : echo_pos]=sonar.convert_cm(sonar.ping_median());
//	Serial.print(echo_pos,DEC);
//	Serial.print('=');
//	Serial.print(distance[(echo_pos==3) ? 1 : echo_pos],DEC);
//	Serial.print(' ');
	echo_pos++;
	echo_pos%=4;
	se.write(((echo_pos==3) ? 1 : echo_pos)*45+45);
}

// Move all servos to midpoints
void park()
{	if (side==s_l)
	{	// Left lifted, so move left legs first
		left_side(l_m,l_m);
		right_lift();
		right_side(r_m,r_m);
		no_lift();
	} else
	if (side==s_r)
	{	// Right lifted, so move right legs first
		right_side(r_m,r_m);
		left_lift();
		left_side(l_m,l_m);
		no_lift();
	} else
	if (side==s_m)
	{	// Neither side lifted, so should already be parked
	}
	if (echo_pos==0 || echo_pos==2)
		echo_scan();
	servo_home();
}

// Walk towards open space
void walk_auto()
{	// Get the initial ranges
	echo_scan();
	delay(100);
	echo_scan();
	delay(100);
	echo_scan();
	delay(100);
	echo_scan();
	if ((distance[0]<60) && (distance[1]<60) && (distance[2]<60))
	{	// Something on all sides within 60cm, so back up
		reverse();
	} else
	if ((distance[1]>150) || ((distance[1]>=distance[0]) && (distance[1]>=distance[2])))
	{	// More than 150 cm in front, or forward is better than left/right
		forward();
	} else
	if (distance[0]<distance[2])
	{	// Right nearer than left
		left();
	} else
	{	// Left nearer than right
		right();
	}
}

void leg_test(void)
{	// Reset positions
	servo_home();
	se.write(60);
	srr.write(r_r);	delay(1000); srr.write(r_m);
	src.write(rc_u);delay(1000); src.write(rc_m);
	srf.write(r_f); delay(1000); srf.write(r_m);
	se.write(120);
	slr.write(l_r);	delay(1000); slr.write(l_m);
	slc.write(lc_u);delay(1000); slc.write(lc_m);
	slf.write(l_f); delay(1000); slf.write(l_m);
	se.write(90);
}

int repeat=1;		// Current repeat count
int repeat_remaining=0;	// Repeats remaining to do

int do_key_action(int key)
{	switch(key)
	{	case 0x0022:	// Left
		case 0x0180:
			echo_scan();
			left();
			return 1;
		case 0x00C2:	// Right
		case 0x01C0:
			echo_scan();
			right();
			return 1;
		case 0x0062:	// Up
		case 0x0158:
			echo_scan();
			forward();
			return 1;
		case 0x00A8:	// Down
		case 0x01A0:
			echo_scan();
			reverse();
			return 1;
		case 0x0002:	// OK
		case 0x0140:
			if (side==s_m)
			{	// Already parked, so sit down
				int const extra=20;
				servo_pos(l_f-extra,lc_m-extra,l_r+extra,r_f+extra,rc_m+extra,r_r-extra);
				delay(1000);
				servo_home();
			} else
			{	park();
			}
			break;
		case 0x0042:	// * - Auto
		case 0x0100:
			walk_auto();
			return 1;
		case 0x0052:	// # - Range find / display + leg test
		case 0x0198:
			{	int distance=sonar.convert_cm(sonar.ping_median());
				Serial.print(distance,DEC);
				Serial.print('\n');
			}
			leg_test();
			break;
		case 0x0068:	// 1
		case 0x01E0:
			repeat=1;
			break;
		case 0x0098:	// 2
		case 0x0110:
			repeat=2;
			break;
		case 0x00B0:	// 3
		case 0x0190:
			repeat=3;
			break;
		case 0x0030:	// 4
		case 0x0150:
			repeat=4;
			break;
		case 0x0018:	// 5
		case 0x01D8:
			repeat=5;
			break;
		case 0x007A:	// 6
		case 0x01F8:
			repeat=6;
			break;
		case 0x0010:	// 7
		case 0x0130:
			repeat=7;
			break;
		case 0x0038:	// 8
		case 0x01B0:
			repeat=8;
			break;
		case 0x005A:	// 9
		case 0x0170:
			repeat=9;
			break;
		case 0x004A:	// 0 - Continuous
		case 0x01F0:
			repeat=10;
			break;
		case 0xFFFF:	// Key held down
			break;
		default:
			Serial.print(key,HEX);
			break;
	}
	return 0;
}

int last_key=0;
int repeat_count=0;

void loop()
{	int key=decode_remote();
	if (key)
	{	last_key=key;
		repeat_count=repeat;
	}
	if (last_key!=0 && repeat_count>0)
	{	if (do_key_action(last_key)==0)
		{	repeat_count=0;
			last_key=0;
		} else
		if (repeat_count<10)
		{	repeat_count--;
		}
	}
}

